from typing import Tuple

import pytest
from google.cloud.bigquery import SchemaField

from gbq.helpers import (
    _check_if_map,
    _map_raw_dictionary_to_bq_schema,
    get_bq_credentials,
    get_bq_schema_from_json_schema,
    get_bq_schema_from_record,
)


@pytest.fixture()
def key_value_schema() -> Tuple:
    return tuple(
        [
            SchemaField("key", "STRING", "NULLABLE"),
            SchemaField("value", "INTEGER", "NULLABLE"),
        ]
    )


@pytest.fixture()
def string_key_value_schema() -> Tuple:
    return tuple(
        [
            SchemaField("key", "STRING", "NULLABLE"),
            SchemaField("value", "STRING", "NULLABLE"),
        ]
    )


def test_get_bq_credentials(mocker):
    expected = '{"secret": "secret"}'
    mocker.patch(
        "gbq.helpers.service_account.Credentials.from_service_account_info"
    ).return_value = '{"secret": "secret"}'
    response = get_bq_credentials('{"secret": "secret"}')
    assert response == expected


def test_convert_to_bq_schema(nested_json_schema):
    address_schema = SchemaField("address", "RECORD", "REPEATED")
    address_nested_schema = [
        SchemaField("id", "INTEGER", "NULLABLE"),
        SchemaField("street", "STRING", "NULLABLE"),
    ]
    address_schema._fields = address_nested_schema
    expected = [
        SchemaField("id", "INTEGER", "NULLABLE"),
        SchemaField("username", "STRING", "NULLABLE"),
        address_schema,
    ]

    response = get_bq_schema_from_json_schema(nested_json_schema)
    assert response == expected


def test__check_if_map_true():
    raw_input = {"5": 5, "10": 10, "11": 11}
    assert _check_if_map(raw_input)


def test__check_if_map_false():
    raw_input = {"request_id": 5, "requested_by": "[unknown]"}
    assert not _check_if_map(raw_input)


def test__map_raw_dictionary_to_bq_schema(string_key_value_schema):
    dates_schema = SchemaField("dates", "RECORD", "REPEATED")
    dates_schema._fields = string_key_value_schema

    data_schema = SchemaField("data", "RECORD", "REPEATED")
    data_nested_schema = tuple(
        [
            SchemaField("id", "INTEGER", "NULLABLE"),
            SchemaField("street", "STRING", "NULLABLE"),
            dates_schema,
        ]
    )
    data_schema._fields = data_nested_schema
    expected = [
        SchemaField("request_id", "INTEGER", "NULLABLE"),
        SchemaField("numbers", "INTEGER", "REPEATED"),
        data_schema,
        SchemaField("requested_by", "STRING", "NULLABLE"),
    ]

    raw_data = {
        "request_id": 1,
        "numbers": [5, 10, 11],
        "data": [
            {
                "id": 1,
                "street": "abc.123",
                "dates": [
                    {"key": "5", "value": "2020-05-11"},
                    {"key": "10", "value": "2020-05-11"},
                ],
            }
        ],
        "requested_by": "user_unknown",
    }

    response = _map_raw_dictionary_to_bq_schema(raw_data)
    assert response == expected


def test_get_bq_schema_from_raw_data_dictionary_list_of_int():
    expected = [
        SchemaField("request_id", "INTEGER", "NULLABLE"),
        SchemaField("numbers", "INTEGER", "REPEATED"),
    ]

    raw_input_data = {
        "request_id": 2,
        "numbers": [1, 2, 3],
    }

    response = get_bq_schema_from_record(raw_input_data)
    assert response == expected


def test_get_bq_schema_from_raw_data_dictionary_list_of_dicts():

    nested_fields = tuple(
        [
            SchemaField("id", "INTEGER", "NULLABLE"),
            SchemaField("value", "FLOAT", "NULLABLE"),
            SchemaField("comment", "STRING", "NULLABLE"),
        ]
    )
    some_data_schema = SchemaField("some_data", "RECORD", "REPEATED")
    some_data_schema._fields = nested_fields

    expected = [
        SchemaField("request_id", "INTEGER", "NULLABLE"),
        some_data_schema,
    ]

    raw_data = {
        "request_id": 2,
        "some_data": [
            {"id": 1, "value": 6.1167402267456055, "comment": "test"},
            {"id": 1, "value": 6.4262800216674805, "comment": "test"},
        ],
    }

    response = get_bq_schema_from_record(raw_data)
    assert response == expected


def test_get_bq_schema_from_raw_data_dictionary_empty_list():
    expected = [
        SchemaField("request_id", "INTEGER", "NULLABLE"),
        SchemaField("numbers", "RECORD", "REPEATED"),
    ]

    raw_data = {
        "request_id": 2,
        "numbers": [],
    }

    response = get_bq_schema_from_record(raw_data)
    assert response == expected


def test_get_bq_schema_from_raw_data_dictionary_maps(
    key_value_schema, string_key_value_schema
):
    repeated_record_schema = SchemaField("repeated_record", "RECORD", "REPEATED")
    repeated_record_schema._fields = key_value_schema

    some_ids_schema = SchemaField("some_ids", "RECORD", "REPEATED")
    some_ids_schema._fields = string_key_value_schema

    some_data = tuple(
        [
            SchemaField("id", "INTEGER", "NULLABLE"),
            SchemaField("value", "FLOAT", "NULLABLE"),
            SchemaField("comment", "STRING", "NULLABLE"),
        ]
    )
    some_data_schema = SchemaField("some_data", "RECORD", "REPEATED")
    some_data_schema._fields = some_data

    expected = [
        SchemaField("request_id", "INTEGER", "NULLABLE"),
        repeated_record_schema,
        some_ids_schema,
    ]

    raw_data = {
        "request_id": 2,
        "repeated_record": {"5": 5, "10": 10, "11": 11},
        "some_ids": {"5": "5", "10": "10", "11": "11"},
    }

    response = get_bq_schema_from_record(raw_data)
    assert response == expected


def test_get_bq_schema_from_raw_data_dictionary_dict():
    some_data_fields = tuple(
        [
            SchemaField("id", "INTEGER", "NULLABLE"),
            SchemaField("value", "FLOAT", "NULLABLE"),
            SchemaField("comment", "STRING", "NULLABLE"),
        ]
    )
    some_data_schema = SchemaField("some_data", "RECORD", "NULLABLE")
    some_data_schema._fields = some_data_fields

    expected = [
        SchemaField("request_id", "INTEGER", "NULLABLE"),
        some_data_schema,
    ]

    raw_data = {
        "request_id": 2,
        "some_data": {
            "id": 1,
            "value": 6.1167402267456055,
            "comment": "test",
        },
    }

    response = get_bq_schema_from_record(raw_data)
    assert response == expected



from enum import Enum
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, Field, root_validator, validator


class StructureType(Enum):
    table = "table"
    view = "view"
    stored_procedure = "stored_procedure"


class PartitionType(Enum):
    time = "time"
    range = "range"


class TimeType(Enum):
    DAY = "DAY"
    HOUR = "HOUR"
    MONTH = "MONTH"
    YEAR = "YEAR"


class BigQueryDataType(Enum):
    TYPE_KIND_UNSPECIFIED = "TYPE_KIND_UNSPECIFIED"
    INT64 = "INT64"
    BOOL = "BOOL"
    FLOAT64 = "FLOAT64"
    STRING = "STRING"
    BYTES = "BYTES"
    TIMESTAMP = "TIMESTAMP"
    DATE = "DATE"
    TIME = "TIME"
    DATETIME = "DATETIME"
    GEOGRAPHY = "GEOGRAPHY"
    NUMERIC = "NUMERIC"
    BIGNUMERIC = "BIGNUMERIC"
    ARRAY = "ARRAY"
    STRUCT = "STRUCT"


class TimeDefinition(BaseModel):
    type: TimeType
    expirationMs: Optional[str]
    field: Optional[str]


class RangeFieldDefinition(BaseModel):
    start: int
    end: int
    interval: int


class RangeDefinition(BaseModel):
    field: str
    range: RangeFieldDefinition


class Partition(BaseModel):
    type: PartitionType
    definition: Union[TimeDefinition, RangeDefinition]


class Argument(BaseModel):
    name: str
    data_type: BigQueryDataType

    @validator("data_type", pre=True)
    def str_or_list_(cls, v):
        return v.upper()


class Structure(BaseModel):
    table_schema: List[Dict] = Field([], alias="schema")
    partition: Union[Partition, None] = None
    clustering: Union[List[str], None] = None
    labels: Dict[str, str] = Field({})
    description: Union[str, None] = None
    view_query: Union[str, None] = None
    body: Union[str, None] = None
    type: Union[StructureType, None] = None
    arguments: Union[List[Argument], None] = None

    @root_validator
    def validate_type(cls, values):
        if not values.get("type", None):
            if values["view_query"]:
                values["type"] = StructureType.view
            elif values["body"]:
                values["type"] = StructureType.stored_procedure
            else:
                values["type"] = StructureType.table
        return values

    @validator("view_query", "body", pre=True)
    def str_or_list_(cls, v):
        if isinstance(v, list) and not [s for s in v if not isinstance(s, str)]:
            v = "\n".join(v)
        return v



from typing import List, Optional

import pytest

from gbq.dto import Structure


class DatasetListItem:
    def __init__(self, dataset_id: str, project: str):
        self.dataset_id = dataset_id
        self.project = project


class SchemaField:
    def __init__(
        self, name: str, mode: str, field_type: str, fields=None, description=None
    ):
        if fields is None:
            fields = []
        self.name = name
        self.mode = mode
        self.field_type = field_type
        self.fields = fields
        self.description = description


class Table:
    def __init__(
        self,
        project: str,
        dataset_id: str,
        table_id: str,
        table_type: str,
        schema: Optional[List[SchemaField]] = None,
        view_query: Optional[str] = None,
        clustering_fields: Optional[list] = None,
        time_partitioning=None,
        range_partitioning=None,
        labels=None,
        description=None,
    ):
        self.project = project
        self.dataset_id = dataset_id
        self.table_id = table_id
        self.table_type = table_type
        self.schema = schema
        self.view_query = view_query
        self.clustering_fields = clustering_fields
        self.time_partitioning = time_partitioning
        self.range_partitioning = range_partitioning
        self.labels = labels
        self.description = description


class Routine:
    def __init__(
        self,
        routine_id: str,
        type_: str,
        language: str,
        body: str,
    ):
        self.routine_id = routine_id
        self.type_ = type_
        self.language = language
        self.body = body


@pytest.fixture()
def nested_json_schema():
    return [
        {"name": "id", "mode": "NULLABLE", "type": "INTEGER"},
        {"name": "username", "mode": "NULLABLE", "type": "STRING"},
        {
            "fields": [
                {"name": "id", "mode": "NULLABLE", "type": "INTEGER"},
                {"name": "street", "mode": "NULLABLE", "type": "STRING"},
            ],
            "name": "address",
            "mode": "REPEATED",
            "type": "RECORD",
        },
    ]


@pytest.fixture()
def nested_json_schema_with_time_partition(nested_json_schema):
    return {
        "schema": nested_json_schema,
        "partition": {"type": "time", "definition": {"type": "DAY"}},
        "type": "table",
    }


@pytest.fixture()
def nested_json_schema_with_incorrect_partition(nested_json_schema):
    return {
        "schema": nested_json_schema,
        "partition": {"type": "time", "definition": {"type": "day"}},
        "type": "table",
    }


@pytest.fixture()
def nested_json_schema_with_clustering(nested_json_schema):
    return {
        "schema": nested_json_schema,
        "clustering": ["id"],
        "type": "table",
    }


@pytest.fixture()
def nested_json_schema_with_partition_and_clustering(nested_json_schema):
    return {
        "schema": nested_json_schema,
        "clustering": ["id"],
        "partition": {"type": "time", "definition": {"type": "DAY"}},
        "type": "table",
    }


@pytest.fixture()
def nested_json_schema_with_labels(nested_json_schema):
    return {
        "schema": nested_json_schema,
        "labels": {"test": "test"},
        "type": "table",
    }


@pytest.fixture()
def nested_json_schema_with_description(nested_json_schema):
    return {
        "schema": nested_json_schema,
        "description": "This is a table",
        "type": "table",
    }


@pytest.fixture()
def dataset_list_items() -> List[DatasetListItem]:
    return [
        DatasetListItem(dataset_id="abc", project="project"),
        DatasetListItem(dataset_id="def", project="project"),
    ]


@pytest.fixture()
def tables() -> List[Table]:
    return [
        Table(project="project", dataset_id="abc", table_id="123", table_type="TABLE"),
        Table(project="project", dataset_id="def", table_id="456", table_type="TABLE"),
    ]


@pytest.fixture()
def table_view() -> List[Table]:
    return [
        Table(project="project", dataset_id="abc", table_id="123", table_type="TABLE"),
        Table(project="project", dataset_id="def", table_id="456", table_type="VIEW"),
    ]


@pytest.fixture()
def views() -> List[Table]:
    return [
        Table(project="project", dataset_id="abc", table_id="123", table_type="VIEW"),
        Table(project="project", dataset_id="def", table_id="456", table_type="VIEW"),
    ]


@pytest.fixture()
def schema_fields() -> List[SchemaField]:
    return [
        SchemaField(name="id", mode="NULLABLE", field_type="INTEGER"),
        SchemaField(name="username", mode="NULLABLE", field_type="STRING"),
    ]


@pytest.fixture()
def nested_schema_fields() -> List[SchemaField]:
    nested_fields = [
        SchemaField(name="id", mode="NULLABLE", field_type="INTEGER"),
        SchemaField(name="username", mode="NULLABLE", field_type="STRING"),
    ]
    return [
        SchemaField(name="id", mode="NULLABLE", field_type="INTEGER"),
        SchemaField(name="street", mode="NULLABLE", field_type="STRING"),
        SchemaField(
            name="address", mode="REPEATED", field_type="RECORD", fields=nested_fields
        ),
    ]


@pytest.fixture()
def table_with_schema(schema_fields) -> Table:
    return Table(
        project="project1",
        dataset_id="abc",
        table_id="123",
        table_type="TABLE",
        schema=schema_fields,
    )


@pytest.fixture()
def sql_schema() -> str:
    return "SELECT * FROM table"


@pytest.fixture()
def table_structure(nested_json_schema) -> Structure:
    return Structure(
        **{
            "schema": nested_json_schema,
            "type": "table",
        }
    )


@pytest.fixture()
def table_structure_with_clustering_partition(
    nested_json_schema_with_partition_and_clustering,
) -> Structure:
    return Structure(**nested_json_schema_with_partition_and_clustering)


@pytest.fixture()
def view_structure(sql_schema) -> Structure:
    return Structure(**{"view_query": sql_schema, "type": "view"})


@pytest.fixture()
def view_structure_with_labels_and_description(sql_schema) -> Structure:
    return Structure(
        **{
            "view_query": sql_schema,
            "labels": {"test": "test"},
            "description": "test",
            "type": "view",
        }
    )


def raw_routine():
    return {"body": "SELECT * FROM table"}


def raw_table():
    return [
        {"name": "id", "mode": "NULLABLE", "type": "INTEGER"},
        {"name": "street", "mode": "NULLABLE", "type": "STRING"},
        {
            "fields": [
                {"name": "id", "mode": "NULLABLE", "type": "INTEGER"},
                {"name": "street", "mode": "NULLABLE", "type": "STRING"},
            ],
            "name": "address",
            "mode": "REPEATED",
            "type": "RECORD",
        },
    ]


@pytest.fixture()
def routine() -> Routine:
    return Routine(
        routine_id="project.dataset.structure",
        type_="PROCEDURE",
        language="SQL",
        body="SELECT * FROM table",
    )


@pytest.fixture()
def routine_structure() -> Structure:
    return empty_arguments_structure()


def empty_arguments_structure() -> Structure:
    return Structure(
        body="SELECT * FROM table",
    )


def structure_with_arguments() -> Structure:
    return Structure(
        body="SELECT * FROM table", arguments=[{"name": "x", "data_type": "DATE"}]
    )



import pytest
from google.api_core.exceptions import NotFound
from google.cloud import bigquery
from google.cloud.bigquery import PartitionRange, QueryJob
from google.cloud.bigquery.routine import RoutineArgument
from pydantic import ValidationError

from gbq.bigquery import BigQuery
from gbq.dto import (
    Partition,
    RangeDefinition,
    RangeFieldDefinition,
    Structure,
    TimeDefinition,
)
from gbq.exceptions import GbqException
from tests.fixtures import (
    Routine,
    Table,
    empty_arguments_structure,
    raw_routine,
    raw_table,
    structure_with_arguments,
)


@pytest.fixture()
def bq(mocker) -> BigQuery:
    mocker.patch(
        "gbq.helpers.service_account.Credentials.from_service_account_info"
    ).return_value = '{"secret": "secret"}'
    mocker.patch("gbq.bigquery.bigquery.Client")
    return BigQuery('{"secret": "secret"}', "project")


@pytest.fixture()
def table(mocker) -> Table:
    table = Table(
        project="project",
        dataset_id="dataset",
        table_id="structure",
        table_type="TABLE",
    )
    mocker.patch(
        "gbq.helpers.service_account.Credentials.from_service_account_info"
    ).return_value = '{"secret": "secret"}'
    mocker.patch("gbq.bigquery.BigQuery._add_partitioning_scheme")
    return table


def test_create_bigquery_class(mocker):
    mocker.patch(
        "gbq.helpers.service_account.Credentials.from_service_account_info"
    ).return_value = '{"secret": "secret"}'
    mocker.patch("gbq.bigquery.bigquery.Client")
    BigQuery('{"secret": "secret"}', "project")
    assert True


def test_get_dataset_in_project(bq, dataset_list_items):
    bq.bq_client.list_datasets.return_value = dataset_list_items
    response = bq.get_dataset_in_project("project")
    assert response == dataset_list_items


def test_get_table_in_project(bq, dataset_list_items, tables):
    bq.bq_client.list_datasets.return_value = dataset_list_items
    bq.bq_client.list_tables.return_value = tables
    response = bq.get_table_in_project("project")
    assert response == tables + tables


def test_get_structure(bq, table_with_schema):
    bq.bq_client.get_table.return_value = table_with_schema
    response = bq.get_structure("project", "dataset", "structure")
    assert response == table_with_schema


@pytest.mark.parametrize(
    "mocked_structure, expected",
    [
        (raw_routine(), Routine),
        (raw_table(), Table),
    ],
)
def test_create_or_update_structure_response(mocker, bq, mocked_structure, expected):
    mocker.patch("gbq.bigquery.BigQuery._handle_stored_procedure")
    mocker.patch("gbq.bigquery.BigQuery._handle_table_or_view")
    bq._handle_stored_procedure.return_value = Routine(
        routine_id="project.dataset.structure",
        type_="PROCEDURE",
        language="SQL",
        body="SELECT * FROM table",
    )
    bq._handle_table_or_view.return_value = Table(
        project="project1",
        dataset_id="abc",
        table_id="123",
        table_type="TABLE",
    )
    response = bq.create_or_update_structure(
        "project", "dataset", "structure", mocked_structure
    )
    assert isinstance(response, expected)


def test_get_routine(bq, routine):
    bq.bq_client.get_routine.return_value = routine
    response = bq.get_routine("project", "dataset", "structure")
    assert response == routine


def test_create_or_update_structure_with_no_partition(bq, nested_json_schema, table):
    bq.create_or_update_structure("project", "dataset", "structure", nested_json_schema)
    bq._add_partitioning_scheme.assert_not_called()
    assert True


def test_create_or_update_structure_with_default_time_partition(
    bq, nested_json_schema_with_time_partition, table
):
    bq.bq_client.get_table.side_effect = NotFound("")
    bq.create_or_update_structure(
        "project", "dataset", "structure", nested_json_schema_with_time_partition
    )

    bq._add_partitioning_scheme.assert_called_once()
    assert True


def test_create_or_update_structure_with_incorrect_partition(
    bq, nested_json_schema_with_incorrect_partition, table
):
    bq.bq_client.get_table.side_effect = NotFound("")
    with pytest.raises(ValidationError):
        bq.create_or_update_structure(
            "project",
            "dataset",
            "structure",
            nested_json_schema_with_incorrect_partition,
        )


def test_create_table_with_clustering(bq, nested_json_schema_with_clustering, table):
    bq.bq_client.get_table.side_effect = NotFound("")
    table = bq.create_or_update_structure(
        "project",
        "dataset",
        "structure",
        nested_json_schema_with_clustering,
    )

    assert table.clustering_fields == nested_json_schema_with_clustering["clustering"]


def test_create_table_without_clustering(bq, nested_json_schema, table):
    bq.bq_client.get_table.side_effect = NotFound("")
    table = bq.create_or_update_structure(
        "project",
        "dataset",
        "structure",
        nested_json_schema,
    )

    assert not table.clustering_fields


def test_update_table_with_clustering(bq, nested_json_schema_with_clustering, table):
    table = bq.create_or_update_structure(
        "project",
        "dataset",
        "structure",
        nested_json_schema_with_clustering,
    )

    assert table.clustering_fields == nested_json_schema_with_clustering["clustering"]


def test_update_table_without_clustering(mocker, bq, nested_json_schema, table):
    mocker.patch(
        "gbq.bigquery.BigQuery.create_or_update_structure"
    ).return_value = table
    response_table = bq.create_or_update_structure(
        "project",
        "dataset",
        "structure",
        nested_json_schema,
    )

    assert not response_table.clustering_fields


def test_create_table_with_labels(bq, nested_json_schema_with_labels, table):
    structure = bq.create_or_update_structure(
        "project",
        "dataset",
        "structure",
        nested_json_schema_with_labels,
    )

    assert structure.labels == nested_json_schema_with_labels["labels"]


def test_create_table_without_labels(mocker, bq, nested_json_schema, table):
    mocker.patch(
        "gbq.bigquery.BigQuery.create_or_update_structure"
    ).return_value = table
    table = bq.create_or_update_structure(
        "project",
        "dataset",
        "structure",
        nested_json_schema,
    )

    assert not table.labels


def test_update_table_with_same_labels(bq, nested_json_schema_with_labels, table):
    table = bq.create_or_update_structure(
        "project",
        "dataset",
        "structure",
        nested_json_schema_with_labels,
    )

    assert table.labels == nested_json_schema_with_labels["labels"]


def test_update_table_without_labels(bq, nested_json_schema_with_labels, table):
    structure = bq.create_or_update_structure(
        "project",
        "dataset",
        "structure",
        nested_json_schema_with_labels,
    )

    assert structure.labels == nested_json_schema_with_labels["labels"]


def test_create_table_with_description(bq, nested_json_schema_with_description, table):
    structure = bq.create_or_update_structure(
        "project",
        "dataset",
        "structure",
        nested_json_schema_with_description,
    )

    assert structure.description == nested_json_schema_with_description["description"]


def test_create_table_without_description(mocker, bq, nested_json_schema, table):
    mocker.patch(
        "gbq.bigquery.BigQuery.create_or_update_structure"
    ).return_value = table
    table = bq.create_or_update_structure(
        "project",
        "dataset",
        "structure",
        nested_json_schema,
    )

    assert not table.description


def test_update_table_with_same_description(
    bq, nested_json_schema_with_description, table
):
    table = bq.create_or_update_structure(
        "project",
        "dataset",
        "structure",
        nested_json_schema_with_description,
    )

    assert table.description == nested_json_schema_with_description["description"]


def test_update_table_without_description(
    bq, nested_json_schema_with_description, table
):
    response_table = bq.create_or_update_structure(
        "project",
        "dataset",
        "structure",
        nested_json_schema_with_description,
    )

    assert (
        response_table.description == nested_json_schema_with_description["description"]
    )


def test__get_table_schema(bq, nested_json_schema):
    expected = Structure(**{"schema": nested_json_schema})
    assert bq._get_structure(nested_json_schema) == expected


def test__get_table_schema_with_partition(bq, nested_json_schema_with_time_partition):
    expected = Structure(**nested_json_schema_with_time_partition)
    assert bq._get_structure(nested_json_schema_with_time_partition) == expected


def test__add_partitioning_scheme_time(bq, table_with_schema):
    time_definition = TimeDefinition(type="DAY")
    partition_scheme = Partition(type="time", definition=time_definition)

    expected = bigquery.TimePartitioning(type_="DAY")
    bq_structure = bq._add_partitioning_scheme(
        bq_structure=table_with_schema, partition_scheme=partition_scheme
    )

    assert not bq_structure.range_partitioning
    assert bq_structure.time_partitioning == expected


def test__add_partitioning_scheme_time_with_field_and_expiration(bq, table_with_schema):
    time_definition = TimeDefinition(type="DAY", expirationMs="1", field="id")
    partition_scheme = Partition(type="time", definition=time_definition)

    expected = bigquery.TimePartitioning(type_="DAY", field="id", expiration_ms="1")
    bq_structure = bq._add_partitioning_scheme(
        bq_structure=table_with_schema, partition_scheme=partition_scheme
    )

    assert not bq_structure.range_partitioning
    assert bq_structure.time_partitioning == expected


def test__add_partitioning_scheme_range(bq, table_with_schema):
    range_field_definition = RangeFieldDefinition(start=1, end=100000, interval=10)
    range_definition = RangeDefinition(field="id", range=range_field_definition)
    partition_scheme = Partition(type="range", definition=range_definition)

    expected = bigquery.RangePartitioning(field=range_definition.field)
    expected.range_ = PartitionRange(**range_definition.range.__dict__)
    bq_structure = bq._add_partitioning_scheme(
        bq_structure=table_with_schema, partition_scheme=partition_scheme
    )

    assert not bq_structure.time_partitioning
    assert bq_structure.range_partitioning.field == expected.field
    assert bq_structure.range_partitioning.range_.start == expected.range_.start
    assert bq_structure.range_partitioning.range_.end == expected.range_.end
    assert bq_structure.range_partitioning.range_.interval == expected.range_.interval


def test__handle_create_structure_table_without_partition_and_clustering(
    bq, table_structure
):
    bq_structure = bq._handle_create_structure(
        dataset="dataset",
        project="project",
        structure_id="structure",
        structure=table_structure,
    )
    assert not bq_structure.clustering_fields
    assert not bq_structure.time_partitioning
    assert not bq_structure.range_partitioning


def test__handle_create_structure_table_with_partition_and_clustering(
    bq, table_structure_with_clustering_partition
):
    bq_structure = bq._handle_create_structure(
        dataset="dataset",
        project="project",
        structure_id="structure",
        structure=table_structure_with_clustering_partition,
    )

    expected_time_partition = bigquery.TimePartitioning(type_="DAY")

    assert bq_structure.clustering_fields == ["id"]
    assert bq_structure.time_partitioning == expected_time_partition
    assert not bq_structure.range_partitioning


def test__handle_create_structure_view(bq, view_structure):
    bq_structure = bq._handle_create_structure(
        dataset="dataset",
        project="project",
        structure_id="structure",
        structure=view_structure,
    )
    assert not bq_structure.clustering_fields
    assert not bq_structure.time_partitioning
    assert not bq_structure.range_partitioning
    assert not bq_structure.schema
    assert not bq_structure.labels
    assert not bq_structure.description
    assert bq_structure.view_query == view_structure.view_query


def test__handle_create_structure_view_with_label_and_description(
    bq, view_structure_with_labels_and_description
):
    bq_structure = bq._handle_create_structure(
        dataset="dataset",
        project="project",
        structure_id="structure",
        structure=view_structure_with_labels_and_description,
    )
    assert not bq_structure.clustering_fields
    assert not bq_structure.time_partitioning
    assert not bq_structure.range_partitioning
    assert not bq_structure.schema
    assert bq_structure.labels == view_structure_with_labels_and_description.labels
    assert (
        bq_structure.description
        == view_structure_with_labels_and_description.description
    )
    assert (
        bq_structure.view_query == view_structure_with_labels_and_description.view_query
    )


@pytest.mark.parametrize(
    "param, expected",
    [
        (empty_arguments_structure(), []),
        (
            structure_with_arguments(),
            [
                RoutineArgument(
                    name="x",
                    data_type=bigquery.standard_sql.StandardSqlDataType(
                        type_kind=bigquery.StandardSqlTypeNames.DATE
                    ),
                )
            ],
        ),
    ],
)
def test__handle_routine_arguments(bq, param, expected):
    response = bq._handle_routine_arguments(param)
    assert response == expected


def test__handle_stored_procedure_create(bq, routine_structure):
    bq.bq_client.get_routine.side_effect = NotFound("")
    bq._handle_stored_procedure("project", "dataset", "structure", routine_structure)
    bq.bq_client.create_routine.assert_called_once()


def test__handle_stored_procedure_update(bq, routine, routine_structure):
    bq.bq_client.get_routine.return_value = routine
    bq._handle_stored_procedure("project", "dataset", "structure", routine_structure)
    bq.bq_client.update_routine.assert_called_once()


def test_delete_dataset(bq):
    bq.bq_client.get_dataset.return_value = True
    response = bq.delete_dataset("project", "dataset")
    assert response


def test_delete_dataset_when_not_exists(bq):
    bq.bq_client.get_dataset.side_effect = NotFound("")
    response = bq.delete_dataset("project", "dataset")
    assert not response


def test_delete_table_or_view(bq):
    bq.bq_client.get_table.return_value = True
    response = bq.delete_table_or_view("project", "dataset", "table")
    assert response


def test_delete_table_or_view_when_not_exists(bq):
    bq.bq_client.get_table.side_effect = NotFound("")
    response = bq.delete_table_or_view("project", "dataset", "table")
    assert not response


def test_execute_raise_exception(bq):
    bq.bq_client.query.side_effect = TypeError

    with pytest.raises(GbqException):
        bq.execute("SELECT * FROM some_table")


def test_execute_raise_no_exception(bq):
    query = "SELECT * FROM some_table"
    bq.bq_client.query.return_value = QueryJob(
        job_id="test_id", query=query, client=bq.bq_client
    )
    response = bq.execute(query=query)

    assert type(response) == QueryJob



from typing import Dict, List, Optional, Union

from google.api_core.exceptions import NotFound
from google.cloud import bigquery
from google.cloud.bigquery import QueryJob
from google.cloud.bigquery.dataset import DatasetListItem
from google.cloud.bigquery.routine import Routine, RoutineArgument
from google.cloud.bigquery.table import PartitionRange, Table

from gbq.dto import (
    Partition,
    PartitionType,
    RangeDefinition,
    Structure,
    StructureType,
    TimeDefinition,
)
from gbq.exceptions import GbqException, InvalidDefinitionException
from gbq.helpers import get_bq_credentials, get_bq_schema_from_json_schema


class BigQuery:
    """
    BigQueryUtil represent a BigQuery Util resource.

    Args:
        svc_account (str):
            Stringified JSON service account value.
        project (Optional[str]):
            Project bound to the operation.
    """

    def __init__(self, svc_account: str, project: Optional[str] = None):
        self.credentials = get_bq_credentials(svc_account)
        self.bq_client = bigquery.Client(credentials=self.credentials, project=project)

    def get_dataset_in_project(self, project: str) -> List[DatasetListItem]:
        """
        Function returns list of DatasetListItem objects of all the datasets in a project.

        Args:
            project (str):
                Project bound to the operation.

        Returns:
            List[DatasetListItem]: A list of object of BigQuery DatasetListItem.
        """
        self.bq_client.project = project
        datasets: List[DatasetListItem] = list(self.bq_client.list_datasets())
        return datasets

    def delete_dataset(self, project: str, dataset: str):
        """
        Function deletes a dataset.

        Args:
            project (str):
                Project bound to the operation.
            dataset (str):
                ID of dataset containing the table.

        Returns:
            Bool: Whether dataset was deleted or not.
        """
        self.bq_client.project = project

        try:
            bq_structure = self.bq_client.get_dataset(dataset)
            self.bq_client.delete_dataset(bq_structure, delete_contents=True)
        except NotFound:
            return False

        return True

    def get_table_in_project(self, project: str) -> List[Table]:
        """
        Function returns list of Table objects of all the tables and views in a project.

        Args:
            project (str):
                Project bound to the operation.

        Returns:
            List[Table]: A list of object of BigQuery Table.
        """
        tables: List[Table] = []
        self.bq_client.project = project
        datasets: List[DatasetListItem] = self.get_dataset_in_project(project)
        for dataset in datasets:
            tables_in_dataset = self.bq_client.list_tables(
                f"{dataset.project}.{dataset.dataset_id}"
            )
            [tables.append(table) for table in tables_in_dataset]  # type: ignore
        return tables

    def get_structure(self, project: str, dataset: str, structure: str) -> Table:
        """
        Function returns a BigQuery Table object.

        Args:
            project (str):
                Project bound to the operation.
            dataset (str):
                ID of dataset containing the table.
            structure (str):
                ID of the structure, can be a table or a view.

        Returns:
            Table: An object of BigQuery Table.
        """
        self.bq_client.project = project
        full_table_name = f"{project}.{dataset}.{structure}"
        bq_table: Table = self.bq_client.get_table(full_table_name)
        return bq_table

    def delete_table_or_view(self, project: str, dataset: str, structure: str):
        """
        Function deletes table or view.

        Args:
            project (str):
                Project bound to the operation.
            dataset (str):
                ID of dataset containing the table.
            structure (str):
                ID of the structure.

        Returns:
            Bool: Whether table or view was deleted or not.
        """
        self.bq_client.project = project

        try:
            bq_structure = self.get_structure(project, dataset, structure)
            self.bq_client.delete_table(bq_structure)
        except NotFound:
            return False

        return True

    def get_routine(self, project: str, dataset: str, routine_name: str) -> Routine:
        """
        Function returns a BigQuery Table object.

        Args:

            project (str):
                Project bound to the operation.
            dataset (str):
                ID of dataset containing the table.
            routine_name (str):
                ID of the structure, can be a table or a view.

        Returns:
            Routine: An object of BigQuery Routine.
        """
        self.bq_client.project = project
        routine_id = f"{project}.{dataset}.{routine_name}"
        routine: Routine = self.bq_client.get_routine(routine_id)
        return routine

    def create_or_update_structure(
        self,
        project: str,
        dataset: str,
        structure_id: str,
        json_schema: Union[List[Dict], Dict],
    ) -> Union[Table, Routine]:
        """
        Function creates/updates provided json schema to the structure.

        Args:
            project (str):
                Project bound to the operation.
            dataset (str):
                ID of dataset containing the structure.
            structure_id (str):
                ID of the structure.
            json_schema (Union[List[Dict], Dict]):
                Raw JSON schema of the table. If List[Dict] it is assumed that the value is
                    table schema, if Dict than value can include partitioning scheme, clustering,
                    table schema, view query, labels, etc

        Examples:
            List[Dict]:
            [
                {
                    "name": "name1",
                    "type: "INTEGER",
                    "mode: "REQUIRED"
                },
                {
                    "name": "name2",
                    "type: "STRING"
                }...
            ]

            Dict:
            List[Dict]:
            {
                "clustering": ["name"],
                "partition": {
                    "type": "time",
                    "definition": {
                        "type": "DAY"
                    }
                },
                "description": "this is a partitioned table",
                "labels": {"team: "abc"},
                "schema: [
                    {
                        "name": "name1",
                        "type: "INTEGER",
                        "mode: "REQUIRED"
                    },
                    {
                        "name": "name2",
                        "type: "STRING"
                    }...
                ]
            }


        Returns:
            Table: An object of BigQuery Table.
        """
        self.bq_client.project = project

        structure = self._get_structure(json_schema)

        if (
            structure.type == StructureType.table
            or structure.type == StructureType.view
        ):
            return self._handle_table_or_view(dataset, project, structure_id, structure)
        elif structure.type == StructureType.stored_procedure:
            return self._handle_stored_procedure(
                dataset, project, structure_id, structure
            )
        else:
            raise InvalidDefinitionException("Missing required structure definition")

    def _handle_table_or_view(
        self, dataset: str, project: str, structure_id: str, structure: Structure
    ):
        """
        Function creates/updates BigQuery Table per the provided information.

        Args:
            project (str):
                Project bound to the operation.
            dataset (str):
                ID of dataset containing the table.
            structure_id (str):
                ID of the routine.
            structure (Structure):
                An object of internal Structure class.

        Returns:
            Routine: An object of BigQuery Routine.
        """
        fields_to_update = []

        try:
            bq_structure = self.get_structure(project, dataset, structure_id)

            if structure.type == StructureType.table:
                fields_to_update.append("schema")
                schema = get_bq_schema_from_json_schema(structure.table_schema)
                bq_structure.schema = schema
            elif structure.type == StructureType.view:
                fields_to_update.append("view_query")
                bq_structure.view_query = structure.view_query

            if structure.labels and structure.labels != bq_structure.labels:
                fields_to_update.append("labels")
                bq_structure.labels = structure.labels

            if (
                structure.description
                and structure.description != bq_structure.description
            ):
                fields_to_update.append("description")
                bq_structure.description = structure.description

            if (
                structure.clustering
                and structure.clustering != bq_structure.clustering_fields
            ):
                fields_to_update.append("clustering")
                bq_structure.clustering_fields = structure.clustering  # type: ignore

            self.bq_client.update_table(bq_structure, fields_to_update)

            return bq_structure
        except NotFound:
            return self._handle_create_structure(
                dataset, project, structure_id, structure
            )

    @staticmethod
    def _get_structure(json_schema: Union[Dict, List[Dict]]) -> Structure:
        """
        Function returns an object of Structure, curated from the input.

        Args:
            json_schema (Union[List[Dict], Dict]):
                Raw JSON schema of the table. If List[Dict] it is assumed that the value is
                    table schema, if Dict than value can include partitioning scheme, clustering
                    and table schema.

        Returns:
            Structure: An object of Structure.
        """
        if isinstance(json_schema, dict):
            return Structure(**json_schema)
        elif isinstance(json_schema, list):
            return Structure(**{"schema": json_schema})

        return  # type: ignore

    def _handle_create_structure(
        self, dataset: str, project: str, structure_id: str, structure: Structure
    ) -> Table:
        """
        Function returns an object of BigQuery Table. It is used for creating new table.

        Args:
            project (str):
                Project bound to the operation.
            dataset (str):
                ID of dataset containing the table.
            structure_id (str):
                ID of the table.
            structure (Structure):
                An object of internal TableSchema class.

        Returns:
            Table: An object of BigQuery Table.
        """
        # Create BQ Table object
        bq_structure = bigquery.Table(f"{project}.{dataset}.{structure_id}")

        if structure.table_schema:
            # Get BQ Schema from JSON provided
            schema = get_bq_schema_from_json_schema(structure.table_schema)
            bq_structure.schema = schema

            # Configure Partition
            if structure.partition:
                bq_structure = self._add_partitioning_scheme(
                    bq_structure=bq_structure, partition_scheme=structure.partition
                )

            # Configure Clustering
            bq_structure.clustering_fields = structure.clustering  # type: ignore

        if structure.view_query:
            bq_structure.view_query = structure.view_query

        if structure.labels:
            bq_structure.labels = structure.labels

        if structure.description:
            bq_structure.description = structure.description

        self.bq_client.create_table(bq_structure)
        return bq_structure

    def _handle_stored_procedure(
        self, dataset: str, project: str, structure_id: str, structure: Structure
    ) -> Routine:
        """
        Function creates/updates BigQuery routine per the provided body and arguments.

        NOTE: Currently we only support PROCEDURE type routines.

        Args:
            project (str):
                Project bound to the operation.
            dataset (str):
                ID of dataset containing the table.
            structure_id (str):
                ID of the routine.
            structure (Structure):
                An object of internal Structure class.

        Returns:
            Routine: An object of BigQuery Routine.
        """
        self.bq_client.project = project
        routine_id = f"{project}.{dataset}.{structure_id}"

        try:
            routine = self.get_routine(project, dataset, structure_id)
            routine.body = structure.body
            routine.arguments = self._handle_routine_arguments(structure)
            routine.description = structure.description

            routine = self.bq_client.update_routine(
                routine,
                [
                    "body",
                    # Due to a limitation of the API,
                    # all fields are required, not just
                    # those that have been updated.
                    "arguments",
                    "language",
                    "type_",
                    "return_type",
                    "description",
                ],
            )
        except NotFound:
            routine = Routine(
                routine_id,
                type_="PROCEDURE",
                language="SQL",
                body=structure.body,
            )
            routine.description = structure.description
            routine.arguments = self._handle_routine_arguments(structure)

            routine = self.bq_client.create_routine(routine)
        return routine

    @staticmethod
    def _handle_routine_arguments(structure: Structure) -> List[RoutineArgument]:
        """
        Function returns list of objects RoutineArgument type if any specified

        Args:
            structure (Structure):
                An object of internal Structure class.

        Returns:
            List[RoutineArgument]: List of objects of RoutineArgument
        """
        bq_arguments = []
        if structure.arguments:
            for argument in structure.arguments:
                arg = RoutineArgument(
                    name=argument.name,
                    data_type=bigquery.StandardSqlDataType(
                        type_kind=bigquery.StandardSqlTypeNames[
                            argument.data_type.value
                        ]
                    ),
                )
                bq_arguments.append(arg)
        return bq_arguments

    def _add_partitioning_scheme(
        self, bq_structure: Table, partition_scheme: Partition
    ) -> Table:
        """
        Function adds partitioning scheme to the table based on the passed configuration.

        Args:
            bq_structure (Table):
                An object of BigQuery Table.
            partition_scheme (Partition):
                An object of internal Partition.

        Returns:
            Table: An object of BigQuery Table.
        """
        if partition_scheme.type.value == PartitionType.time.value:
            time_partitioning = self._get_time_partitioned_scheme(partition_scheme)
            bq_structure.time_partitioning = time_partitioning
        elif partition_scheme.type.value == PartitionType.range.value:
            range_partitioning = self._get_range_partitioned_scheme(partition_scheme)
            bq_structure.range_partitioning = range_partitioning

        return bq_structure

    @staticmethod
    def _get_time_partitioned_scheme(
        partition_scheme: Partition,
    ) -> bigquery.TimePartitioning:
        """
        Function returns a BQ Time Partitioning scheme from Partition object.

        Args:
            partition_scheme (Partition):
                An object of internal Partition.

        Returns:
            bigquery.TimePartitioning: An object of bigquery.TimePartitioning.
        """
        definition: TimeDefinition = partition_scheme.definition  # type: ignore
        time_partitioning = bigquery.TimePartitioning(type_=definition.type.value)
        if definition.field:
            time_partitioning.field = definition.field
        if definition.expirationMs:
            time_partitioning.expiration_ms = definition.expirationMs
        return time_partitioning

    @staticmethod
    def _get_range_partitioned_scheme(
        partition_scheme: Partition,
    ) -> bigquery.RangePartitioning:
        """
        Function returns a BQ Range Partitioning scheme from Partition object.

        Args:
            partition_scheme (Partition):
                An object of internal Partition.

        Returns:
            bigquery.RangePartitioning: An object of bigquery.TimePartitioning.
        """
        definition: RangeDefinition = partition_scheme.definition  # type: ignore
        range_partitioning = bigquery.RangePartitioning(field=definition.field)
        range_partitioning.range_ = PartitionRange(**definition.range.__dict__)
        return range_partitioning

    def execute(self, query: str) -> QueryJob:
        """
        Function return a QueryJob object after executing a SQL statement

        Args:
            query (str):
                BigQuery query string

        Returns:
            QueryJob
                An object of QueryJob.
        """
        try:
            query_job = self.bq_client.query(query)

            # Wait for query job to finish.
            query_job.result()

            return query_job
        except Exception as e:
            raise GbqException(str(e)) from e



#!/usr/bin/env python

# We still need a setup.py shim so we can still pip install -e .
# https://snarky.ca/what-the-heck-is-pyproject-toml/#how-to-use-pyproject-toml-with-setuptools

import setuptools

if __name__ == "__main__":
    setuptools.setup()
